--[[
-- Copyright 2016 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- This module defines a page for selecting values for components (like
-- clothing or PC personalizations like gender). The screen is laid out
-- as follows:
--
--  (F1) Search Box
--  (F2) component one
--  (F3) component two
--  ...
--  (F12)
--
--  Typing in the search box narrows the list of components down to those
--  that match the search string. The game automatically selects the
--  first choice in the list, so one can type in a string and then
--  press enter if the one they want is displayed.
--
--  To track this, we use the following fields on uistate:
--  creation - A table that maps component ids to the currently
--      selected value for that component. This will be modified by
--      this screen, and is used by the character creation screen to
--      update the player if they accept the changes.
--  visiblevalues - An ordered table containing only the values that
--  match the player's search string. The table is keyed on the value's
--  index in component.values so that the search is consistent. There are two entries:
--      id: The index of the value in the list component.values, where
--          component is the component we're currently looking at
--      value: The value itself, this is typically a string, but may
--          be something more elaborate in the case of say clothing
--  search - A table containing a single key "text" which contains the player's
--      search string
--  selected - The index from visiblevalues that is currently selected
--]]


local fonts = require 'ui/fonts'
local pages = require 'pages'

local M = {}

local function displayable(item, slot)
    return item.slots[1] == slot and item.selectable
end

local function buildgui(uistate, component)
    local gui = uistate.gui
    local options = uistate.core.gamestate.options
    local buttonboldfont = fonts.buttonBold(options.buttonfont)
    local buttonfont  = fonts.button(options.buttonfont)
    local options = uistate.core.gamestate.options
    local componentsHeight = buttonboldfont:getHeight()
    
    local titlefont = fonts.title(options.titlefont)
    local titlefontoptions = {font=titlefont}
    gui.Label(
        "Number/Search:",
        titlefontoptions,
        gui.layout:col(
            titlefont:getWidth("Number/Search:") + 5,
            titlefont:getHeight()
        )
    )
    if not uistate.search then
        uistate.search = {text=""}
    end

    local inputoptions = {font=titlefont, id="search"}
    local placeholder = string.rep(" ", 20)
    local placeholderw = titlefont:getWidth(placeholder) + 5
    local w = math.max(
        placeholderw,
        titlefont:getWidth(uistate.search.text) + 5
    )
    local h = titlefont:getHeight()
    local x,y = gui.layout:col(w, h)

    if not uistate.visiblevalues then
        uistate.visiblevalues = {_keys={}}
        for i, value in ipairs(component.values) do
            uistate.visiblevalues[i] = value
            table.insert(uistate.visiblevalues._keys, i)
        end
    end

    gui.Input(uistate.search, inputoptions, x, y, w, h)
    local maxWidth = 0
    for key, value in pairs(uistate.visiblevalues) do
        local valuestring = type(value) == "string" and value or value.name
        if key ~= "_keys" then
            maxWidth = math.max(maxWidth, titlefont:getWidth(valuestring) + 20)
        end
    end
    for _, i in ipairs(uistate.visiblevalues._keys) do
        local value = uistate.visiblevalues[i]
        local valuestring = (type(value) == "string" and value or value.name)
        local label = tostring(i) .. ") "
            .. valuestring
        local font = uistate.selected == i and buttonboldfont or buttonfont
        gui.Button(
            label,
            {id=valuestring, font=font, align="left"},
            gui.layout:row(font:getWidth(label) + 5, font:getHeight())
        )
        local scarlettables = require 'scarlettables'
        if value.description then
            gui.layout:push(gui.layout:nextCol()) 
            local labelid = uistate.visiblevalues[i].name .. "_description"
            local scarlettables = require 'scarlettables'
            local description = component.values[i].description
            gui.Label(
                description,
                {id=labelid, font=font, align="left"},
                gui.layout:col(font:getWidth(description) + 20, font:getHeight())
            )
            gui.layout:pop()
        end
    end

    if not gui.hasKeyboardFocus(inputoptions.id) then
        gui.setHit(inputoptions.id)
        gui.grabKeyboardFocus(inputoptions.id)
    end

    local searchstring = string.gsub(uistate.search.text, "%-", "%%-")
    searchstring = string.gsub(searchstring, "%(", "%%(")
    searchstring = string.gsub(searchstring, "%)", "%%)")
    searchstring = ".*" .. searchstring .. ".*"
    local oldvisiblecount = #uistate.visiblevalues._keys
    uistate.visiblevalues = {_keys={}}
    for i, value in ipairs(component.values) do
        -- If we have an exact match, just show that value.
        if uistate.search.text == value then
            uistate.visiblevalues = {[i]=value, _keys={i}}
            break
        end
        -- Allows us to search based on the value's index
        local pattern = tostring(i) 
            .. (type(value) == "string" and value or value.name)
        if #searchstring == 0 or string.match(pattern, searchstring) then
            uistate.visiblevalues[i] = value
            table.insert(uistate.visiblevalues._keys, i)
        end
    end
    -- If the count changed we modified the search string somehow.
    if not uistate.selected or oldvisiblecount ~= #uistate.visiblevalues._keys then
        _, uistate.selected = next(uistate.visiblevalues._keys)
    end
    local confirmlabel = "(Enter) Confirm"
    gui.Button(
        confirmlabel,
        {id="return", font=buttonfont, align="center"},
        gui.layout:row(titlefont:getWidth(confirmlabel) + 5, buttonfont:getHeight())
    )
    gui.Button(
        "(Esc) Exit",
        {id="escape", font=buttonfont, align="center"},
        gui.layout:col()
    )
end

--[[
-- This registers the selected value for the component with the
-- uistate. It populates a table `uistate.creation` that maps
-- the component's id to the selected value. This is used by the
-- creation screen to display the currently selected value
--
-- The type of value depends on what is being displayed. For
-- example, if it's clothing, then value will be the table
-- used to model clothing. If the component is a simple
-- PC personalization option like gender which is just a
-- string, then value is a string
--
-- uistate: The current state of the game
-- component: The personalization component whose value is
-- currently being changed
--]]
local function registerselection(uistate, component)
    if not uistate.creation then
        uistate.creation = {}
    end
    local value = uistate.visiblevalues[uistate.selected]
    uistate.creation[component.id] = type(value) == "string" 
        and value 
        or value.name
    uistate.search.text = ""
    uistate.selected = nil
    uistate.visiblevalues = nil
end

--[[
-- Takes a string representing a value for a component, and
-- returns a page that applies that value to the player.
--
-- valueindex: The index in uistate.visiblevalues for the value to apply
-- applychangespage: The page that applies a change after its
-- been selected
--
-- return A page that applies the value to the player
--]]
local function selectandapply(valueindex, applychangespage)
    return pages.page(
        nil,
        function() return applychangespage end,
        nil,
        function(uistate) uistate.selected = valueindex end
    )
end

--[[
-- Defines the choices for the component page. (Enter) applies
-- the currently selection, while (Esc) cancels.
--
-- uistate: The current state of the UI
-- component: The component we're selecting a value for
-- previouspage: The page to go back to instead of applying changes
-- page: This page
--
-- returns An ordered table of choices available on this page
--]]
local function componentchoices(uistate, component, previouspage, page)
    local applychangespage = pages.page(
        nil,
        function() return previouspage end,
        nil,
        function(uistate) registerselection(uistate, component) end
    )
    local choices = {
        ["return"]=applychangespage,
        escape=pages.page(
            nil, 
            function() return previouspage end,
            nil,
            function(uistate) 
                uistate.search.text = "" 
                uistate.visiblevalues = nil
            end
        ),
        _keys={"return", "escape"}
    }
    for i, value in ipairs(component.values) do
        local valuestring = type(value) == "string" and value or value.name
        choices[valuestring] = selectandapply(i, applychangespage),
        table.insert(choices._keys, valuestring)
    end
    choices.up = pages.page(
        nil,
        function() return page end,
        nil,
        function(uistate)
            local keys = uistate.visiblevalues._keys
            local selectedindex = 1
            for i, index in ipairs(keys) do
                if uistate.selected == index then
                    selectedindex = i
                end
            end
            uistate.selected = uistate.selected == keys[1]
                and keys[#keys]
                or keys[selectedindex - 1]
        end
    )
    choices.down = pages.page(
        nil,
        function() return page end,
        nil,
        function(uistate)
            local keys = uistate.visiblevalues._keys
            local selectedindex = 1
            for i, index in ipairs(keys) do
                if uistate.selected == index then
                    selectedindex = i
                end
            end
            uistate.selected = uistate.selected == keys[#keys]
                and keys[1]
                or keys[selectedindex + 1]
        end
    )
    return choices
end

--[[
-- Returns a page for displaying this component
--
-- component: Models the component whose values are being selected from. A
-- table with the following keys:
--  id - The id to use for the GUI widgit that displays this component in the
--      character creation/status screen.
--  label - The label to display to the player for this component
--  values - A list of the possible values that this component can take. 
--      This is either a list of Strings (i.e. the component values can influence
--      game text, but don't provide any gameplay value), or a table with two
--      keys: 
--          name - the name of the value, displayed to the user
--          description - a description of the value.
-- returns a UI page that displays the component and interacts with user input
--]]

function M.page(component, previouspage)
    local page = pages.page()
    function page.choices(uistate)
        return componentchoices(uistate, component, previouspage, page)
    end
    page.drawgui = function(uistate) buildgui(uistate, component) end
    return page
end
return M
