--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

local M = {}
local combatstateops = require 'combat/combatstate'
local F = require 'f-strings/F'
local log = require 'log'
local pages = require 'pages'
local scarlettables = require 'scarlettables'
local utilities = require 'combat/combatutils'

--[[
-- Given the attack primitive, and the current state of combat, returns
-- the damage that the user will do against the target, and a single paragraph
-- describing the attack.
--
-- This is very similar to the M.defaultdamage, except it doesn't trigger
-- retaliation.
--
-- primitive: The damage primitive
-- combatstate: The current state of combat
--
-- return The damage done, nil for retaliation, and a list of strings 
-- describing the action
--]]
function M.shootdamage(user, target, combatstate)
    local attack = M.attackstat(user)
    local defense = M.defensestat(target)
    local damagedone = M.shield(target, math.max(0, attack - defense))
    local text = F"{user.combatname} shoots {target.combatname} for {damagedone} damage!"
    return damagedone, nil, {text}
end

--[[
-- Resolves the given LOVE KeyConstants into targets.
--
--  combatstate: The current state of combat
--  ...: A variable number of KeyConstants to be resolved into targets
--
-- returns A list of characters to be targeted, or nil and an optional
-- error message if a target cannot be resolved
-- ]]
function M.resolveattacktargets(combatstate, ...)
    local targets = {}
    for _, targetkey in ipairs({...}) do
        assert(targetkey, "Targetkey must be non-nil.")
        log.info("-----------Attacking-------------")
        log.info("Targetkey: " .. targetkey)
        local target = utilities.resolvetargetkey(combatstate, targetkey)
        log.info(
            (target and target.name or "nil")
            .. " targeted for attack."
        )
        if not target then
            log.info("Unknown target " .. targetkey)
            log.info("-----------End Attacking-------------")
            return nil
        end
        if combatstateops.isdefeated(combatstate, target.name) then
            log.info("Target already defeated: " .. target.name)
            log.info("-----------End Attacking-------------")
            return nil, target.combatname .. " is already defeated!"
        end
        table.insert(targets, target)
    end
    return targets
end

--[[
-- Given a variable number of keys pressed to select a target, the current state of
-- combat, and the page to go to in case of an error, returns the next combat
-- page to execute.
--
-- If the player presses a key that does not map to a target, then the previous
-- page is returned directly.
--
-- combatstate - The current stat of combat
-- previouspage - The page to transition to if the target is invalid. A target
--  is invalid if:
--      1. They aren't active (i.e. they've been defeated)
--      2. There is no target with the specified key
-- damage - A function that takes the attacker, target, and current state of
--  combat, and returns the amount of damage to do, any retaliatory damage
--  to do, and a description of the
--  action as a list of strings, one entry per paragraph. Defaults to a function
--  that uses the atacker's attack and shield (for blocking retaliatory damage), 
--  and target's defense, shield, and retaliation.
--  then modifies it based on target's relative speed
-- ... - One or more Love KeyConstants the player pressed to select a target
--
-- return - A page that does damage to each of the targets, or the previouspage
--  if we failed to resolve at least one of the targets
-- ]]
function M.attack(combatstate, previouspage, damage, ...)
    local user = combatstate.characters[combatstate.charindex]
    local targets, errormsg = M.resolveattacktargets(combatstate, ...)
    if not targets then
        if errormsg then
            return pages.page(
                nil,
                function() return previouspage end,
                nil,
                function(combatstate)
                    table.insert(combatstate.combatlog, errormsg)
                end
            )
        else
            return previouspage
        end
    end
    return M.attacktargets(targets, user, damage)
end

--[[
-- Returns a combat page that damages the specified targets using the
-- specified damage function
--
-- targets - The targets to be attacked
-- user - The attacker
-- damage - A function that takes the attacker, target, and current state of
--  combat, and returns the amount of damage to do, any retaliatory damage
--  to do, and a description of the
--  action as a list of strings, one entry per paragraph. Defaults to a function
--  that uses the atacker's attack and shield (for blocking retaliatory damage), 
--  and target's defense, shield, and retaliation.
--  then modifies it based on target's relative speed
--]]
function M.attacktargets(targets, user, damage)
    damage = damage or M.defaultdamage
    return pages.page(
        nil,
        nil,
        nil,
        function(combatstate)
            M.damage(combatstate, damage, targets, user)
        end
    )
end

--[[
-- Damages the specified targets by the amount determined by the passed in 
-- function.
--
-- combatstate - The current stat of combat
-- damage - A function that takes the attacker, target, and current state of
--  combat, and returns the amount of damage to do, any retaliatory damage
--  to do, and a description of the
--  action as a list of strings, one entry per paragraph. Defaults to a function
--  that uses the atacker's attack and shield (for blocking retaliatory damage), 
--  and target's defense, shield, and retaliation.
-- targets - The list of characters to damage
-- user - The character doing the damage
-- ]]
function M.damage(combatstate, damage, targets, user)
    damage = damage or M.defaultdamage
    for _, target in ipairs(targets) do
        local damage, retaliation, description = damage(
            user, 
            target, 
            combatstate
        )
        local tenergy = target.statistics.energy
        target.statistics.energy = math.max(0, tenergy - damage)
        local uenergy = user.statistics.energy
        if retaliation and target.statistics.energy > 0 then
            user.statistics.energy = math.max(0, uenergy - retaliation)
            table.insert(
                description, 
                target.combatname .. " retaliates for " 
                .. tostring(retaliation) .. " damage!"
            )
        end
        for _, paragraph in ipairs(description) do
            table.insert(combatstate.combatlog, paragraph)
        end
        if target.statistics.energy <= 0 then
            table.insert(combatstate.combatlog, F"{target.combatname} collapses!")
            combatstateops.defeatcharacter(combatstate, target.name)
        end
        if user.statistics.energy <= 0 then
            table.insert(combatstate.combatlog, F"{user.combatname} collapses!")
            combatstateops.defeatcharacter(combatstate, user.name)
        end
    end
end

-- [[
-- Identical to M.attack, except that it uses M.shootdamage to calculate damage,
-- rather than the default damage formula. M.shootdamage is identical to default
-- damage, except it doesn't trigger retaliation.
-- ]]
function M.shoot(combatstate, previouspage, ...)
    return M.attack(combatstate, previouspage, M.shootdamage, ...)
end

function M.applymodifiers(basestat, modifiers) 
    local modifiedstat = basestat
    for _, modifier in pairs(modifiers) do
        modifiedstat = modifiedstat + modifier
    end
    return math.max(0, modifiedstat)
end

--[[
-- Computes the character's attack, using the specified primary statistic
-- as its base stat.
--
-- primarystatistic: The statistic to primarily to compute the attack, defaults
-- to "strength"
--
-- return The attack for this character after all modifiers have been applied,
-- using the specified statistic as its primary value
--]]
function M.attackstat(character, primarystatistic)
    primarystatistic = primarystatistic or "strength"
    local strength = M.statistic(character, primarystatistic)
    local speed = M.statistic(character, "speed")
    local modifiers = character.secondarystatistics.attackmodifiers
    return M.applymodifiers(strength + math.floor(speed/2), modifiers)
end

--[[
-- Computes the character's defense, using speed as its base stat.
--
-- Note that speed only grants half of its value to defense (the other half goes
-- to attack).
--
-- character: The character whose defense stat is desired
--]]
function M.defensestat(character)
    local defensemodifiers = character.secondarystatistics.defensemodifiers
    return M.applymodifiers(M.statistic(character, "speed"), defensemodifiers)
end

function M.statistic(character, statistic)
    local modifiers = character.secondarystatistics[statistic .. "modifiers"]
    return M.applymodifiers(character.statistics[statistic], modifiers)
end

--[[
-- Given a character and some damage to apply to that character, reduces the
-- character's shield by an amount equal to damage, and reduces the damage by
-- an amount equal to shield, and returns the remaining damage (or zero, 
-- whichever is greater).
--
-- If damage is nil, this function instead returns the current value of 
-- shield after applying all modifiers.
--
-- NOTE: If a damage value is provided, then this function has side effects:
-- it reduces the character's shield.
--
-- character: The character whose shield is reducing the damage
-- damage: The damage reducing (and being reduced by) the character's shield
--
-- returns The remaining damage to apply to the character after being reduced
-- by their shield, or the current strength of the character's shield if 
-- no damage is provided
--]]
function M.shield(character, damage)
    local shieldmodifiers = character.secondarystatistics.shield
    local shield = M.applymodifiers(0, shieldmodifiers)
    if not damage then
        return shield
    end
    if shield > damage then
        shieldmodifiers.damage = shieldmodifiers.damage - damage
    else
        character.secondarystatistics.shield = {damage=0}
    end
    return math.max(0, damage - shield)
end

--[[
-- Computes retalation from the user attacking the target.
--
-- If only one parameter is provided, then the function instead
-- returns the retaliation value of the provided character.
--
-- user: The character attacking, whose getting retaliated against
-- target: The target whose retaliation is triggering
-- damage: The amount of damage to deal
--
-- returns the amount of damage retalation ends up doing after being
-- reduced by shield
--]]
function M.retaliation(user, target, damage)
    if not target then
        target = user
    end
    local retaliationmodifiers = target.secondarystatistics.retaliation
    local retaliation = M.applymodifiers(0, retaliationmodifiers)
    if not damage then
        return retaliation
    end
    if retaliation > 0 then
        retaliationmodifiers.damage = retaliationmodifiers.damage - damage
        return M.shield(user, math.min(retaliation, damage))
    else
        target.secondarystatistics.retaliation = {damage=0}
        return 0
    end
end

local function defaultdamageformula(attacker, defender)
    return math.max(1, M.attackstat(attacker) - M.defensestat(defender))
end

--[[
-- Given the attacker, target, and the current state of combat, returns
-- the damage that the user will do against the target, any retaliatory
-- damage the target does to the user, and a list of paragraphs
-- describing the attack.
--
-- THIS HAS SIDE-EFFECTS. It reduces the target's shield and retaliatory
-- stats, and the user's shield stat based on the base damage done.
--
-- attacker: The character attacking
-- target: The target of the attack
-- combatstate: The current state of combat
-- formula: A function that takes the user and target, and returns the 
-- raw damage to do, before shield is applied. Defaults to 
-- math.max(0, attacker.attack - defender.defense)
--
-- Returns the damage the attacker does to the target, any retaliatory damage
-- the target does to the attacker, and a list of strings 
-- (one entry per paragraph) describing the attack.
--]]
function M.defaultdamage(attacker, target, combatstate, formula)
    formula = formula or defaultdamageformula
    local u = attacker
    local t = target
    local damagedone = M.shield(t, formula(u, t))
    local originalretaliation = M.retaliation(t)
    local retaliatorydamage = originalretaliation > 0 
        and M.retaliation(u, t, damagedone)
        or nil

    local otk
    for _, status in pairs(t.statuses) do
        if status.status.name == "OTK" then
            otk = true
        end
    end
    local text
    -- Special text if the attack is actually a spanking
    if otk then
        -- luacheck: ignore 211
        local uname = u.combatname
        local tname = t.combatname
        local thisher = t.attributes.gender == "female" and "her" or "him"
        local topponents = combatstateops.opponents(combatstate, t)
        local possibletext = {
            "{uname} leans over and lands {damagedone} smacks on {tname} "
                .. "raised, vulnerable bottom.",
            "{uname} rains {damagedone} hard, fast smacks down {tname}'s "
                .. "squirming, jiggling bottom.",
            "{uname} puts a hand on the small of {tname}'s back and peppers "
                .. "{tname}'s upthrust, helpless bottom {damagedone} times.",
            "{tname} kicks, wiggles and yelps while {uname} beats {thisher} "
            .. "round bottom {damagedone} times."
        }
        if #topponents > 1 then
            for _, opponent in pairs(topponents) do
                if opponent ~= u then
                    table.insert(
                        possibletext,
                        F("{opponent.combatname} grabs {tname}'s legs and holds "
                        .. "them still "
                        .. " while {uname} smacks {tname}'s bucking bottom "
                        .. " {damagedone} times.")
                    )
                    break
                end
            end
        end
        text = F(possibletext[math.random(#possibletext)])
    else
        text = F"{u.combatname} attacks {t.combatname} for {damagedone} damage!"
    end
    return damagedone, retaliatorydamage, {text}
end

return M
