--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--[[
-- This module provides hooks for user-custom combat logic. In particular,
-- it provides the function `firstpage` that provides the first choice made
-- in combat. It could be a manual choice (ask for the first command for a
-- character), or an automatic choice (the enemy goes first, or some sort of
-- initializion choice).
--
-- Note that firstpage is invoked at the start of every combat round.
-- ]]
local M = {}

local ai = require 'combat/ai'
local attacks = require 'combat/attack'
local combatoperations = require 'combat/operations'
local combatstateops = require 'combat/combatstate'
local F = require 'f-strings/F'
local log = require 'log'
local pages = require 'pages'
local scarlettables = require 'scarlettables'
local statuses = require 'combat/statuses'


--[[
-- Given the current state of combat, returns the very first page to use for
-- combat.
--
-- The very first page sets up some mappings needed for combat in the
-- combatstate, and then delegates to the first character's choice.
--
-- In particular, this page establishes the `combatstate.choicecounts` table,
-- which tracks the choices each character has left to make this round.
--
-- combatstate - The starting state of combat
--
-- returns The first choice to execute in combat
-- ]]
function M.firstpage(combatstate)
    local activecharacterindex = 1
    local startcharacter = combatstate.characters[activecharacterindex]
    while startcharacter and combatstateops.isdefeated(combatstate, startcharacter.name) do
      activecharacterindex = activecharacterindex + 1
      startcharacter = combatstate.characters[activecharacterindex]
    end
    -- If we don't have any characters alive, then it's a tie, and we just use
    -- the first character.
    local start = startcharacter 
        and startcharacter.choice 
        or combatstate.characters[1].choice
    return pages.page(
        nil,
        function() return start end,
        nil,
        function(combatstate1)
            local activecharacter = combatstateops.activecharacter(combatstate1)
            statuses.startturn(activecharacter, combatstate1)
        end
    )
end

--[[
-- Returns a combat page that finishes up
-- the current round, and starts the next round. At the end of the round,
-- the following occurs:
--  1. The round counter is incremented.
--  2. The previous combat state is dropped
--  3. The combat state's round-long action counts are cleared
--  4. The combatstate's charindex is reset to 1, allowing us to start the
--      next round over
--  4. Statuses are decremented by one (and cured if their duration is reduced
--      to 0)
--  5. Each character loses energy by attrition determined by an `attrition`
--      function added to gamestate by the attrition option
--      (see data/options.lua). The attrition function takes the character,
--      and the combatstate and returns the amount of attrition damage the
--      character should take
--
-- returns A combat page that ends the round and starts the next
--]]
function M.endround()
    local function transformer(combatstate)
        local endofroundmessages = {}
        combatstate.endofroundmessages = endofroundmessages
        combatoperations.nextround(combatstate)
        for _, character in pairs(combatstate.characters) do
            if not combatstateops.isdefeated(combatstate, character.name) then
                statuses.decrementstatuses(character, combatstate, 1)
                M.applyattrition(character, combatstate, endofroundmessages)
            end
            if attacks.shield(character) > 0 then
                local damage = character.secondarystatistics.shield.damage
                character.secondarystatistics.shield.damage = damage - 1
            end
            if attacks.retaliation(nil, character) > 0 then
                local damage = character.secondarystatistics.retaliation.damage
                character.secondarystatistics.retaliation.damage = damage - 1
            end
        end
        local separators = string.rep("-", 15)
        table.insert(
            endofroundmessages,
            separators .. " Round " .. combatstate.round .. " " .. separators
        )
    end
    local function body(combatstate)
        return combatstate.endofroundmessages
    end

    return pages.page(body, M.firstpage, nil, transformer)
end

function M.applyattrition(character, combatstate, endofroundmessages)
    local combatstatemodule = require 'combat/combatstate'
    local energy = character.statistics.energy
    local attrition = math.min(energy - 1, character.attrition)
    character.statistics.energy = math.max(1, energy - attrition)
    if attrition ~= 0 then
        table.insert(
            endofroundmessages,
            F"{character.combatname} takes {attrition} points of "
                .. "attrition damage!"
        )
    end
end

local function basechoicetransformer(combatstate)
    log.info("Clearing metalog.")
    combatstate.metalog = {}
end

local function basechoiceautomatics(combatstate)
    local character = combatstate.characters[combatstate.charindex]
    local choicecounts = combatstate.choicecounts[character.name]
    for _, name in ipairs(choicecounts._keys) do
        local choice = choicecounts[name]
        if choice.current > 0 then
            return not character.pc and ai.aichoice
        end
    end
    -- Since combat automatically goes on to the next character if
    -- a page doesn't have any choices or automatics, an empty page
    -- effectively transitions to the next character.
    return pages.page()
end

local ALPHABET = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l',
    'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'x', 'y', 'z'}


--[[
-- A function that takes the name of a combat choice, the list of keys that
-- are already used for choices, and the string to display to the user to
-- select this choice.
--
-- The returned string is identical to choicename, except the first character in
-- choicename that is not in choicekeys is wrapped in parenthesis. If no such
-- character is found, this function returns nil. Otherwise, it returns the
-- new selection string, and a new list of choicekeys with the character for
-- this selection added to the end.
--
-- choicename - The name of the choice to be turned into a selection string
-- choicekeys - A table from already used choice key to the associated selection
--
-- return The new selection string, and the key used for this selection, or nil if there 
-- are no characters that can be used
--]]
local function buildselection(choicename, choicekeys)
    for key, choice in pairs(choicekeys) do
        if string.find(choice, choicename) then
            return choice, key
        end
    end
    for _, letter in ipairs(ALPHABET) do
        if not choicekeys[letter] then
            choicename = "(" .. letter .. ")" .. choicename
            return choicename, letter
        end
    end
end

--[[
-- Given the current state of combat, builds the table of combat choices
-- available to the currently active character.
--
-- combatstate - The current state of combat
--
-- return An ordered table that maps choices to the combat pages that execute
-- those choices
-- ]]
local function basechoicechoices(combatstate)
    local character = combatstateops.activecharacter(combatstate)
    local choicecounts = combatstate.choicecounts[character.name]
    local combatchoices = {_keys={}}
    local choicekeys = {
        e="end",
        l="load",
        t="titlescreen",
        r="restart"
    }
    for _, name in ipairs(choicecounts._keys) do
        local choice = choicecounts[name]
        local selection
        selection, letter = buildselection(name, choicekeys)
        choicekeys[letter] = selection
        if not selection then
            local message = "All characters in "
                .. " the alphabet have been taken: "
                .. scarlettables.string(combatchoices)
            log.error(message)
            assert(false, message)
        end
        local choicecount = choice.current
        if choice.current > 0 then
            local selectionstring = selection .. ": " .. choice.current
            assert(
                choice.page,
                "Missing a page for " .. scarlettables.string(choice)
            )
            combatchoices[selectionstring] = M.decrementcount(choice.page, name)
            table.insert(combatchoices._keys, selectionstring)
        end
    end
    return combatchoices
end

--[[
-- Returns the standard combat choice.
--]]
function M.basechoice()
    return pages.page(
        nil,
        basechoiceautomatics,
        basechoicechoices,
        basechoicetransformer
    )
end

--[[
-- Returns a page that decrements the active character's count for the
-- specified action type, and then transitions to the specified action.
--
-- actionpage - The action that should be transitioned to after decrementing
--  the count
--  actiontype - The type of the action being taken
-- ]]
function M.decrementcount(actionpage, actiontype)
    return pages.page(
        nil,
        function() return actionpage end,
        nil,
        function(combatstate)
            log.info("Decrementing count for : " .. actiontype)
            local character = combatstate.characters[combatstate.charindex]
            local choice = combatstate.choicecounts[character.name][actiontype]
            choice.current = math.max(0, choice.current - 1)
        end
    )
end

return M
