--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- Defines a page that contains a list of buttons describing the
-- utility commands available during combat.
--]]
local M = {}

local fonts = LOVE and require 'ui/fonts'
local pages = require 'pages'
local scarlettables = require 'scarlettables'

local function drawgui(uistate, uichoices)
    local nextcharacter = "(E)nd Turn"
    local loadtext = "(L)oad"
    local titletext = "(T)itle"
    local restarttext = "(R)estart"
    local increasefont = "(+) Increase combat status font"
    local decreasefont = "(-) Decrease combat status font"
    local concise = "(Tab)Concise Combat Status"
    local returntocombat = "(<==)Return to Combat"
    local maxlength = 0
    local buttontext = {
        nextcharacter,
        loadtext,
        titletext,
        restarttext,
        increasefont,
        decreasefont,
        concise,
        returntocombat
    }

    local gui = uistate.gui
    local buttonfont = fonts.button(uistate.core.gamestate.options.buttonfont)
    for _, text in ipairs(buttontext) do
        local textlength = fonts.width(buttonfont, text)
        if textlength > maxlength then
            maxlength = textlength
        end
    end
    maxlength = maxlength + 5
    local buttonheight = buttonfont:getHeight()
    local x, y = gui.layout:row(maxlength, buttonfont:getHeight())

    gui.Button(
        nextcharacter,
        {id="e", font=buttonfont},
        gui.layout:row(maxlength, buttonheight)
    )

    gui.Button(
        restarttext,
        {id="r", font=buttonfont},
        gui.layout:row(maxlength, buttonheight)
    )

    gui.Button(
        loadtext,
        {id="l", font=buttonfont},
        gui.layout:row(maxlength, buttonheight)
    )

    gui.Button(
        titletext,
        {id="t", font=buttonfont},
        gui.layout:row(maxlength, buttonheight)
    )

    gui.Button(
        increasefont,
        {id="+", font=buttonfont},
        gui.layout:row(maxlength, buttonheight)
    )
    gui.Button(
        decreasefont,
        {id="-", font=buttonfont},
        gui.layout:row(maxlength, buttonheight)
    )
    gui.Button(
        concise,
        {id="tab", font=buttonfont},
        gui.layout:row(maxlength, buttonheight)
    )

    local titlefont = fonts.title(uistate.core.gamestate.options.titlefont)
    gui.Button(
        returntocombat,
        {id="backspace", font=buttonfont},
        gui.layout:row(maxlength, titlefont:getHeight())
    )
end

function M.commands(uichoices, uipage)
    local choicescopy = {_keys={}}
    for _, key in ipairs(uichoices._keys) do
        choicescopy[key] = uichoices[key]
        table.insert(choicescopy._keys, key)
    end
    choicescopy.backspace = uipage
    table.insert(choicescopy._keys, "backspace")
    local commandpage = pages.page(
        nil,
        nil,
        function() return choicescopy end,
        nil
    )

    function commandpage.drawgui(uistate)
        drawgui(uistate, choicescopy)
    end
    return commandpage
end

local function drawactionsgui(uistate, actions, choicescopy)
    local font = fonts.button(uistate.core.gamestate.options.buttonfont)
    local longestwidth = 0
    for _, key in ipairs(actions._keys) do
        local action = actions[key]
        local width = font:getWidth(action)
        if longestwidth < width then
            longestwidth = width + 5
        end
    end
    local height = font:getHeight()
    local gui = uistate.gui
    for _, key in ipairs(actions._keys) do
        local action = actions[key]
        gui.Button(
            action,
            {font=font, id=key},
            gui.layout:row(longestwidth, height)
        )
    end
end

function M.actions(uichoices, actions, uipage)
    local choicescopy = scarlettables.copy(uichoices)
    local actionscopy = scarlettables.copy(actions)
    choicescopy.backspace = uipage
    table.insert(choicescopy._keys, "backspace")
    actionscopy.backspace = "<== Return to Combat Screen"
    table.insert(actionscopy._keys, "backspace")

    local actionspage = pages.page(
        nil,
        nil,
        function() return choicescopy end,
        nil
    )

    function actionspage.drawgui(uistate)
        drawactionsgui(uistate, actionscopy, choicescopy)
    end

    return actionspage
end
return M
