--[[
-- Copyright 2019 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
--[[
-- This module defines the AI's used by the Princess and her minions.
--]]

local M = {}

local ai = require 'combat/ai'
local combatstateops = require 'combat/combatstate'
local log = require 'log'
local pages = require 'pages'
local powerutils = require 'data/powerutils'
local statuses =  require 'data/statuses'
local statusops = require 'combat/statuses'

local function toysbuffed(allies)
    for _, ally in ipairs(allies) do
        if statusops.statusduration(ally, statuses.strongerfaster) then
            return true
        end
    end
    return false
end

local function mrcuddlesinjured(allies)
    local characters = require 'data/characters'
    for _, ally in ipairs(allies) do
        if ally.combatname == characters.mrcuddles.combatname then
            return ally.statistics.energy / ally.statistics.maxenergy <= 0.25
        end
    end
    return false
end

local function findmrcuddles(combatstate)
    for _, character in pairs(combatstateops.activeallies(combatstate)) do
        if character.combatname == "Mr. Cuddles" then
            return character
        end
    end
end

--[[
-- Princess' starting AI. Princess has the following patterns:
-- 0. If Princess isn't protected by Mr. Cuddles, invoke Bearshield.
-- 1. If there are fewer than four villains, summon toys.
-- 2. Otherwise, if none of the toys are buffed, buff all the toys.
-- 3. If Mr. Cuddles' energy is 25% or less of its maximum, use healbear.
-- 4. Otherwise, use bearswipe.
--]]
function M.princess(combatstate)
    local user = combatstateops.activecharacter(combatstate)
    local choicecounts = combatstate.choicecounts[user.name]
    if choicecounts.power.current <= 0 then
        return ai.dumb(combatstate)
    end
    local allies = combatstateops.activeallies(combatstate)
    local mrcuddles = findmrcuddles(combatstate)
    local scarlettables = require 'scarlettables'
    if mrcuddles and not statusops.statusduration(user, statuses.bearguard) then
        return ai.decrementcount(user.powers.bearshield, user, "power")
    elseif #allies < 6 then
        return ai.decrementcount(user.powers.summontoys, user, "power")
    elseif not toysbuffed(allies) then
        log.info("Buffing toys.")
        return ai.decrementcount(user.powers.strengthentoys, user, "power")
    elseif mrcuddlesinjured(allies) then
        log.info("Healing Mr. Cuddles")
        return ai.decrementcount(user.powers.healbear, user, "power")
    else
        return ai.decrementcount(user.powers.bearswipe, user, "power")
    end
end

local function grabbedopponent(opponents)
    for _, opponent in ipairs(opponents) do
        if statusops.statusduration(opponent, statuses.grabbed) then
            return opponent
        end
    end
    return opponent
end

--[[
-- Mr. Cuddles' starting AI. He has the following pattern:
-- 1. If Mr. Cuddles isn't countering, he prepares his counter.
-- He alternates between:
-- 2. If at least one target is grabbed, he gives them a spanking.
-- 3. Otherwise, he grabs someone.
-- and:
-- 4. He attacks a target with bearpunch.
--]]
function M.mrcuddles(combatstate)
    local user = combatstateops.activecharacter(combatstate)
    local countering = statusops.statusduration(user, statuses.counter)
    local choicecounts = combatstate.choicecounts[user.name]
    if choicecounts.power.current > 0 then
        if not countering then
            return ai.decrementcount(user.powers.bearcounter, user, "power")
        elseif combatstate.round % 2 == 0 then
            local opponents = combatstateops.activeopponents(combatstate)
            local grabbed = grabbedopponent(opponents)
            return ai.decrementcount(
                grabbed and user.powers.bearspank or user.powers.beargrab,
                user,
                "power"
            )
        end
        return ai.decrementcount(user.powers.bearpunch, user, "power")
    end
    return ai.dumb(combatstate)
end

-- [[
-- The AI for Sir Stuffins. Sir Stuffins follows the following pattern:
-- 1. If Giraffalot is rearing, and neither of them is away then charge.
-- 2. Alternate between spanking and whacking a target.
-- ]]
function M.sirstuffins(combatstate)
    local user = combatstateops.activecharacter(combatstate)
    local characters = require 'data/characters'
    local giraffalot = combatstateops.findcharacterbycombatname(
        combatstate,
        characters.stuffed_giraffe.combatname
    )
    local choicecounts = combatstate.choicecounts[user.name]
    if choicecounts.power.current > 0 then
        local away = statusops.statusduration(user, statuses.away)
            or statusops.statusduration(giraffalot, statuses.away)
        if statusops.statusduration(giraffalot, "fast") and not away then
            return ai.decrementcount(user.powers.charge, user, "power")
        elseif combatstate.round % 2 == 0 then
            return ai.decrementcount(user.powers["noodle_spank"], user, "power")
        end
        return ai.decrementcount(user.powers["noodle_whack"], user, "power")
    end
    return ai.dumb(combatstate)
end

--[[
-- AI for Giraffalot. Giraffalot has the following pattern:
--
-- 1. Every four turns, rear if neither she nor Sir Stuffins are away.
-- Alternate between:
-- 2. Pinning a target.
-- 3. Trampling a target
--]]
function M.giraffalot(combatstate)
    local user = combatstateops.activecharacter(combatstate)
    local choicecounts = combatstate.choicecounts[user.name]
    local characters = require 'data/characters'
    local sirstuffins = combatstateops.findcharacterbycombatname(
        combatstate,
        characters.stuffed_bear.combatname
    )
    if choicecounts.power.current > 0 then
        local away = statusops.statusduration(user, statuses.away)
            or statusops.statusduration(sirstuffins, statuses.away)
        if combatstate.round % 3 == 0 and not away then
            return ai.decrementcount(user.powers.rear, user, "power")
        elseif combatstate.round % 2 == 0 then
            return ai.decrementcount(user.powers.trample, user, "power")
        end
        return ai.decrementcount(user.powers.pin, user, "power")
    end
    return ai.dumb(combatstate)
end

-- Should be used by Legostopheles when deciding who to spank.
local function paddletarget(opponents, combatstate)
    local bestbufftarget
    local bestbuffvalue = -999
    for _, opponent in ipairs(opponents) do
        if statusops.statusduration(opponent, statuses.grabbed) then
            local bestbuff = statusops.bestbuff(opponent)
            local buffvalue = bestbuff and bestbuff.buff or bestbuffvalue
            if type(buffvalue) == "function" then
                buffvalue = buffvalue(opponent)
            end
            if not bestbufftarget then
                bestbufftarget = opponent
                bestbuffvalue = buffvalue
            elseif bestbuff and buffvalue > bestbuffvalue then
                bestbufftarget = opponent
                bestbuffvalue = buffvalue
            end
        end
    end
    return bestbufftarget
end

local function performspanking(combatstate, user)
    local opponents = combatstateops.activeopponents(combatstate)
    local target = paddletarget(opponents, combatstate)
    local descriptions = {
        string.format(
            "%s flips %s over and whaps %s repeatedly with a lego paddle. %s kicks and howls as the "
            .. "the ridged legos dig into %s %s skin.",
            user.combatname,
            target.combatname,
            powerutils.himher(target),
            target.combatname,
            powerutils.hisher(target),
            powerutils.baring(target) and "soft, exposed" or "soft"
        ),
        string.format(
            "%s holds %s upside down and whaps %s %s bottom like a pinata. %s swings back and "
            .. "forth with every swat, squirming and yelping.",
            user.combatname,
            target.combatname,
            powerutils.hisher(target),
            powerutils.baring(target) and "naked" or "vulnerable",
            target.combatname
        )
    }
    local spankpage = pages.page(
        nil,
        nil,
        nil,
        function(cs)
            local index = math.random(#descriptions)
            table.insert(cs.combatlog, descriptions[index])
            statusops.addstatus(target, statuses.spanked, 1, user, cs)
        end
    )
    return ai.decrementcount(spankpage, user, "spank")
end

--[[
-- Base AI for Legostopheles. Legostopheles follows the following pattern:
-- 1. Use legoswirl to weaken a target if there are any
--  unweakened targets, and Legostopheles' attack count is greater than 5.
-- 3. If there is a target that isn't grabbed, then alternate between grabbing
-- a target and using lego paddle until every target is grabbed and/or
-- Legostopheles can spank twice, or Legostopheles' attack count is 1.
-- 4. If Legostopheles' spank count is less than 2, buff with lego paddle.
-- 5. If not every opponent is grabbed, grab an opponent
--
--  After using all powers, spank the grabbed opponent with the best buff.
--]]
function M.legostopheles(combatstate)
    local user = combatstateops.activecharacter(combatstate)
    local choicecounts = combatstate.choicecounts[user.name]
    local opponents = combatstateops.activeopponents(combatstate)
    local grabbed = statusops.affectedopponent(opponents, statuses.grabbed)
    if choicecounts.power.current > 0 then
        local canswirl = choicecounts.attack.max > 5
            and statusops.unaffectedopponent(opponents, statuses.strong(-1))
            and combatstate.round % 2 == 1
        local cangrab = choicecounts.attack.max > 1
            and statusops.unaffectedopponent(opponents, statuses.grabbed)
        local canspank = (not choicecounts.spank or choicecounts.spank.max < 2)
            and grabbed
        if canswirl then
            return ai.decrementcount(user.powers.legoswirl, user, "power")
        elseif cangrab then
            return ai.decrementcount(user.powers.grab, user, "power")
        elseif canspank then
            return ai.decrementcount(user.powers.legopaddle, user, "power")
        end
    end
    if choicecounts.spank and choicecounts.spank.current > 0 and grabbed then
        return ai.decrementcount(performspanking(combatstate, user), user, "spank")
    end
    return ai.dumb(combatstate)
end

return M
