--[[
-- Copyright 2018 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
-- ]]
--
--
-- [[
-- Here we define Scarlet Moon's powers. All of Scarlet Moon's powers are
-- self-buffs. These buffs do everything, from increasing stats, to shifting
-- stats around, to changing what and how many actions she has. As a result,
-- Scarlet Moon has the potential to be the single most powerful character
-- in the game, but it takes a bit of time for her to fully power up.
--
-- Scarlet Moon's powers have an additional piece of metadata: type. These
-- types impose some restrictions on what buffs Scarlet Moon can have active
-- at any one time. For example, Scarlet Moon can have only one weapon active
-- at one time (though there is a buff that increases the number of active
-- weapons by 1).
-- ]]

local M = {}
local attacks = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local F = require 'f-strings/F'
local pages = require 'pages'
local powers = require 'combat/powers'
local statuses = require 'data/statuses'
local statusops = require 'combat/statuses'

local function hisher(target)
    return target.attributes.gender == "female" and "her" or "his"
end

-- [[
-- Given the current state of combat, the buff to apply, and a 
-- list of strings describing the buff being applied, applies
-- the specified buff to the active character, and updates the
-- combatstate's combatlog with the description.
--
-- combatstate: The current state of combat
-- buff: The buff to apply
-- description: string describing the buff being applied
--
-- return A chained primitive combat page that applies the buff for
-- a duration equal to the active character's power
-- ]]
local function buffscarletmoon(buff, description, combatstate, duration)
    local character = combatstateops.activecharacter(combatstate)
    duration = duration or attacks.statistic(character, "power")
    local character = combatstate.characters[combatstate.charindex]
    table.insert(combatstate.combatlog, description)
    statusops.addstatus(character, buff, duration, character, nil, combatstate)
end

--[[
-- This allows Scarlet Moon to conjure a spring to use in battle. The spring
-- adds the "Away" status to Scarlet Moon's attacks.
--]]
M["s-spring"] = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local description = character.combatname 
            .. " weaves "
            .. hisher(character)
            .. " forcefield into a large scarlet spring."
        buffscarletmoon(statuses.scarletspring, description, combatstate)
        powers.applycost(character, M["s-spring"])
    end
)

M["s-spring"].cost = 2
M["s-spring"].numtargets = 0
M["s-spring"].stats = {"power"}
M["s-spring"].description = "Scarlet Moon's regular attacks "
    .. "inflict the Away"
    .. " status. The Away status reduces the target's attack count by 1. "
    .. "Scarlet Moon can only have 1 weapon active at a time.\n"
    .. "  Type: Weapon"

-- [[
-- Adds the "shoot" action to Scarlet Moon's actions. Shoot is very similar to
-- attack, except that it uses power for its base damage, rather than strength.
-- ]]
M["s-beam"] = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local description = character.combatname 
            .. "'s eyes flare with scarlet light."
        local duration = math.max(10, attacks.statistic(character, "power"))
        local scarletbeam = statuses.scarletbeam
        buffscarletmoon(scarletbeam, description, combatstate, duration)
        powers.applycost(character, M["s-beam"])
    end
)

M["s-beam"].cost = 2
M["s-beam"].numtargets = 0
M["s-beam"].stats = {"power"}
M["s-beam"].description = "Scarlet Moon gains a new Shoot action. Shoot is like "
    .. "Attack (including relying on your attack rating), except it doesn't trigger retaliation. "
    .. "It increases attrition by 1, and lasts as long as Scarlet Moon's power, "
    .. " or 10 turns, whichever is greater."

M["s-armor"] = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local description = character.combatname .. "'s forcefield darkens."
        local duration = attacks.statistic(character, "power")
        table.insert(combatstate.combatlog, description)
        statusops.addstatus(character, statuses["s-armor"], duration, character, nil, combatstate)
        powers.applycost(character, M["s-armor"])
    end
)
M["s-armor"].cost = 2
M["s-armor"].numtargets = 0
M["s-armor"].stats = {"power"}
M["s-armor"].description = function(character)
    local shield = 2 * attacks.statistic(character, "power")
    return F"Gives Scarlet Moon a *{shield}* point shield. "
        .. "Shield potency increases with power."
end

----------------------- Statistic Transfers -------------------------------
local function transferstats(character, sources, target, amount, name)
    local secondarystatistics = character.secondarystatistics
    for _, statname in ipairs(sources) do
        secondarystatistics[statname .. "modifiers"][name] = -amount
    end
    secondarystatistics[target .. "modifiers"][name] = 2*amount
end

local function transferonestat(character, source, targets, amount, transfername)
    character.secondarystatistics[source .. "modifiers"][transfername] = -amount
    local amountperstat = math.max(1, math.floor(amount / #targets))
    for _, name in ipairs(targets) do
        local key = name .. "modifiers"
        character.secondarystatistics[key][transfername] = amountperstat
    end
end

local function cancelstattransfer(combatstate, choicename, transfername)
    local character = combatstateops.activecharacter(combatstate)
    local description = character.combatname 
        .. " cancels "
        .. transfername 
        .. "!"
    table.insert(combatstate.combatlog, description)
    for _, statistic in ipairs(character.statistics._keys) do
        local key = statistic .. "modifiers"
        character.secondarystatistics[key][transfername] = nil
    end
    local choices = combatstate.choicecounts[character.name]
    choices[choicename] = nil
    local choiceindex = 0
    for i, name in ipairs(choices._keys) do
        if name == choicename then
            choiceindex = i
            break
        end
    end
    if choiceindex > 0 then
        table.remove(choices._keys, choiceindex)
    end
end

local function cancelstatus(choices, choicename, transfername)
    local function cancel(combatstate)
        cancelstattransfer(combatstate, choicename, transfername)
    end
    local page = pages.page(nil, nil, nil, cancel)
    choices[choicename] = {
        max=1,
        current=1,
        page=page
    }
    table.insert(choices._keys, choicename)
end

M["s-fists"] = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local combatchoices = combatstate.choicecounts[character.name]
        if combatchoices["cancel s-fists"] then
            table.insert(combatstate.combatlog, "Cannot apply s-fists twice!")
            return
        end
        if combatchoices["cancel s-eyes"] then
            cancelstattransfer(combatstate, "cancel s-eyes", "s-eyes")
        end
        table.insert(
            combatstate.combatlog,
            character.combatname .. "'s fists flare with scarlet light."
        )
        local amount = 4
        local targetstats = {"strength", "speed"}
        transferonestat(character, "power", targetstats, amount, "s-fists")
        cancelstatus(combatchoices, "cancel s-fists", "s-fists")
        powers.applycost(character, M["s-fists"])
    end
)

M["s-fists"].cost = 0
M["s-fists"].numtargets = 0
M["s-fists"].stats = {"power"}
M["s-fists"].description = function(character)
    local transfer = 4
    return F"Transfer {transfer} points from power into speed and strength."
end

M["s-eyes"] = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local combatchoices = combatstate.choicecounts[character.name]
        if combatchoices["cancel s-eyes"] then
            table.insert(combatstate.combatlog, "Cannot apply s-eyes twice!")
            return
        end
        if combatchoices["cancel s-fists"] then
            cancelstattransfer(combatstate, "cancel s-fists", "s-fists")
        end
        table.insert(
            combatstate.combatlog,
            character.combatname .. "'s eyes flare with crimson light."
        )
        local amount = 2
        local sourcestats = {"strength", "speed"}
        transferstats(character, sourcestats, "power", amount, "s-eyes")
        cancelstatus(combatchoices, "cancel s-eyes", "s-eyes")
        powers.applycost(character, M["s-eyes"])
    end
)

M["s-eyes"].cost = 0
M["s-eyes"].numtargets = 0
M["s-eyes"].stats = {"power"}
M["s-eyes"].description = function(character)
    local transfer = 2
    return F"Transfer {transfer} points from speed and strength into power."
        .. " Costs 0 energy"
end

M["s-staff"] = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local description = character.combatname
            .. " lengthens "
            .. hisher(character)
            .. " forcefield into a staff."
        buffscarletmoon(statuses.scarletstaff, description, combatstate)
        powers.applycost(character, M["s-staff"])
    end
)
M["s-staff"].cost = 1
M["s-staff"].numtargets = 0
M["s-staff"].stats = {"power"}
M["s-staff"].description = function(character)
    local numtargets = math.ceil(attacks.statistic(character, "power") / 5) + 1
    return "Sets the number of "
    .. F" targets you can attack to {numtargets}. Increases "
    .. " attrition by 1. Target count increases every 5 points in power."
    .. "Scarlet Moon can only have 1 weapon active at a time.\n"
    .. "  Type: Weapon"
end

M["s-cords"] = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local description = character.combatname
            .. " attaches a glowing cord to something nearby and zips "
            .. "into the air."
        buffscarletmoon(statuses.swinging, description, combatstate)
        powers.applycost(character, M["s-cords"])
    end
)
M["s-cords"].cost = 1
M["s-cords"].numtargets = 0
M["s-cords"].stats = {"power"}
M["s-cords"].description = function(character)
    local increase = 2
    return F"Increase attack and defense by {increase}, and cures Away at"
    .. " the start of your turn. Increases attrition by 2."
end

M["s-paddle"] = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local description = character.combatname
            .. " weaves " .. hisher(character) .. " forcefield into a "
            .. "pingpong paddle."
        buffscarletmoon(statuses.scarletpaddle, description, combatstate)
        powers.applycost(character, M["s-paddle"])
    end
)

M["s-paddle"].cost = 1
M["s-paddle"].numtargets = 0
M["s-paddle"].stats = {"strength", "power"}
M["s-paddle"].description = function(character)
    local spankerstrength = math.floor(attacks.statistic(character, "strength") / 2)
    return "Increase Spank "
        .. "count by 1, reduces Attack count by 1. "
        .. "Spank inflicts Spanked for 1 round. Spanked "
        .. F"may interfere with the powers of some enemies."
        .. F"It also reduces the duration of their "
        .. F"best buff by an amount based on Scarlet Moon's strength, and the.. "
        .. "spankees power.\nType: Weapon"
        .. "1 weapon can be active at a time.\n"
end

M["s-focus"] = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local description = character.combatname 
            .. " takes a deep breath and focuses."
        buffscarletmoon(statuses.scarletfocus, description, combatstate)
        powers.applycost(character, M["s-focus"])
    end
)

M["s-focus"].cost = 3
M["s-focus"].numtargets = 0
M["s-focus"].stats = {"power"}
M["s-focus"].description =
    "Increases Power count by 1, and decreases Attack count by 1."

return M
