--[[
-- Copyright 2018 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
-- ]]
--
-- [[
-- All the statuses that are specific to Scarlet Moon (i.e. the player
-- character). Scarlet Moon's statuses are all self-buffs, though some may
-- come with certain drawbacks.
--
-- Scarlet Moon's statuses include an optional key: type. This type is used
-- to impose restrictions on which buffs Scarlet Moon can have active at any
-- one time. For example, by default Scarlet Moon can only have 1 weapon buff
-- active at a time.
-- ]]
local M = {}
local attacking = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local log = require 'log'
local pages = require 'pages'
local scarletchoices = require 'data/scarletchoices'
local scarlettables = require 'scarlettables'
local statusops = require 'combat/statuses'
local targeting = require 'combat/target'

local function hisher(character)
    return character.attributes.gender == "female" and "her" or "his"
end

-- [[
-- Removes the specified choice from the provided ordered table of choices.
--
-- choices: The ordered table of choices to remove a choice from
-- choicename: The name of the choice to remove
-- ]]
local function removechoice(choices, choicename)
    choices[choicename] = nil
    scarlettables.remove(choices._keys, choicename)
end

--[[
-- Adds a choice to the user that cancels a status they have inflicted on the
-- target.
--
-- choices: The choices to be augmented
-- choicename: The name of the choice to be added
-- status: The status to be canceled by the new choice
-- username: The name of the character canceling the status
--]]
local function cancelstatuscommand(choices, choicename, status, username)
    local function cancel(combatstate)
        local character = combatstateops.findcharacter(combatstate, username)
        statusops.removestatus(character, status, combatstate)
        local description = character.combatname 
            .. " cancels "
            .. status.name 
            .. "!"
        table.insert(combatstate.combatlog, description)
    end
    local page = pages.page(nil, nil, nil, cancel)
    choices[choicename] = {
        max=1,
        current=1,
        page=page
    }
    table.insert(choices._keys, choicename)
end

--[[
-- This function enforces the limitations on weapons. If the character is
-- already inflicted with the newstatus, then the duration of the new status
-- is added to the duration of the old.
--
-- Otherwise, if the character already has their max allowed weapon statuses,
-- then the oldest weapon is removed.
--
-- combatstate - The current state of combat
-- character - The character receiving the weapon buff
-- weapon - The weapon (i.e. status) to give the character
-- duration - How long the character should have the weapon
--
-- returns true if the character already has this weapon, false otherwise
--]]
function M.stackweapons(combatstate, character, weapon)
    local firstweapon
    local numweapons = 0
    for _, statusduration in ipairs(character.statuses) do
        local status = statusduration.status
        if status.name == weapon.name then
            return true
        elseif status.type == "weapon" then
            if not firstweapon then
                firstweapon = statusduration
            end
            numweapons = numweapons + 1
        end
    end
    log.info("Num weapons: " .. tostring(numweapons))
    local maxweaponcount = character.maxweaponcount or 1
    log.info("maxweaponcount: " .. tostring(maxweaponcount))
    local removefirstweapon = numweapons >= maxweaponcount
        and weapon.name ~= firstweapon.status.name
    log.info("first weapon: " .. (firstweapon and firstweapon.status.name or "nil"))
    if removefirstweapon then
        statusops.removestatus(character, firstweapon.status, combatstate)
    end
    return false
end

-- ------------------------ Scarlet Spring -----------------------------------
-- [[
-- Scarlet Spring adds an inflict action to the target's base attacks, allowing
-- them to reduce their enemy's attack count by 1.
-- ]]

M.scarletspring = {
    name="Scarlet Spring",
    type="weapon",
    display="spring",
    concisename="SP",
    buff=1
}

--[[
-- Does damage and applies the Away status. 
--
-- combatstate - The current state of combat
-- previouspage - The page to transition to if we fail to resolve a target
-- damage - A function that takes this primitive and the current state of
--  combat, and returns the amount of damage to do and a description of the
--  action as a list of strings, one entry per paragraph. Defaults to a
--  function that uses strength as the base stat, but
        table.remove(character.statuses, i)
--  then modifies it based on target's relative speed
-- ... - One or more Love KeyConstants the player pressed to select a target
--]]
local function springattack(combatstate, previouspage, ...)
    local attack = attacking.attack(combatstate, previouspage, nil, ...)
    local targets = attacking.resolveattacktargets(combatstate, ...) or {}
    local targetenergy = {}
    for _, target in ipairs(targets) do
        table.insert(targetenergy, target.statistics.energy)
    end
    local basetransformer = attack.transformer
    local statuses = require 'data/statuses'
    local away = statuses.away
    function attack.transformer(combatstate1)
        basetransformer(combatstate1)
        for i, target in ipairs(targets) do
            statusops.addstatus(target, away, 1, target, 1, combatstate1)
            table.insert(
                combatstate1.combatlog, 
                target.combatname .. " is sent flying 100 feet away!"
            )
        end
    end
    return attack
end

--[[
-- Defines the apply operation for the scarletspring status. Scarlet spring adds
-- a modifier that transforms all damage primitives from the specified character
-- into a chain of two primitives, one increases the damage done, while the
-- other inflicts the away status.
--
-- combatstate: The current state of combat
-- character: The character to apply scarletspring to
-- duration: How long the status should last
--]]
function M.scarletspring.apply(character, combatstate)
    local hasweapon = M.stackweapons(combatstate, character, M.scarletspring)
    if hasweapon then
        return
    end
    statusops.increaseattrition(combatstate, character, 1)
    combatstate.choicecounts[character.name].attack.page = targeting.target(
        springattack,
        1
    )
    cancelstatuscommand(
        combatstate.choicecounts[character.name],
        "end spring",
        M.scarletspring,
        character.name
    )
end

--[[
-- Removes the scarlet spring status from the specified character.
--
-- combatstate: The current state of combat
-- character: The character to remove the modifier from
--]]
function M.scarletspring.remove(character, combatstate)
    local baseattack = character.choicecounts.attack.page
    combatstate.choicecounts[character.name].attack.page = baseattack
    statusops.increaseattrition(combatstate, character, -1)
    removechoice(combatstate.choicecounts[character.name], "end spring")
end

------------------------ Scarlet Staff --------------------------------
-- [[
-- Scarlet staff increases the number of enemies Scarlet Moon can target with
-- her attack by ceil(power/5) + 1.
-- ]]
M.scarletstaff = {
    name="Scarlet Staff",
    type="weapon",
    display="staff",
    concisename="ST",
    buff=1
}

--[[
-- Removes the scarlet staff status from the specified character.
--
-- combatstate: The current state of combat
-- character: The character to remove the modifier from
--]]
function M.scarletstaff.remove(character, combatstate)
    removechoice(combatstate.choicecounts[character.name], "end staff")
    combatstate.choicecounts[character.name].attack.page =
        character.choicecounts.attack.page
end

function M.scarletstaff.apply(character, combatstate)
    local hasweapon = M.stackweapons(combatstate, character, M.scarletstaff)
    if hasweapon then
        return
    end
    local function attack(combatstate1, previouspage, ...)
        return attacking.attack(combatstate1, previouspage, nil, ...)
    end
    local numtargets = math.ceil(attacking.statistic(character, "power")/5) + 1
    combatstate.choicecounts[character.name].attack.page = targeting.target(
        attack,
        numtargets
    )
    cancelstatuscommand(
        combatstate.choicecounts[character.name],
        "end staff",
        M.scarletstaff,
        character.name
    )
end

----------------------- Scarlet Beam ----------------------------
-- [[
-- Scarlet Beam adds the Shoot command to the target's command list, allowing
-- them hit an enemy without triggering retaliation.
-- ]]
M.scarletbeam = {
    name="Scarlet Beam",
    display="beam",
    concisename="BE",
    buff=2
}

function M.scarletbeam.apply(character, combatstate)
    local choices = combatstate.choicecounts[character.name]
    statusops.increasechoice(choices, "shoot", 1, scarletchoices.shoot)
    statusops.increaseattrition(combatstate, character, 1)
end

function M.scarletbeam.remove(character, combatstate)
    local choices = combatstate.choicecounts[character.name]
    if choices.shoot then
        removechoice(choices, "shoot")
    end
    statusops.increaseattrition(combatstate, character, -1)
end

----------------------- Swinging ----------------------------
-- [[
-- Swinging increases your character's attack and defense by power/2 and 
-- increases attrition by 2. It also cancels any Away you may be inflicted 
-- with at the start of your turn
-- ]]
M.swinging = {
    name="Swinging",
    display="swing",
    concisename="SW",
    buff=3
}

-- [[
-- Applies the swinging status to the given character.
--
-- Swinging applies all the changes documented above. In order to ensure
-- that it can correctly reverse the bonus to speed, it saves the bonus to
-- combatstate.statusstate[M.swinging.name][character.name], creating
-- combatstate.statusstate[M.swinging.name] in the process if it doesn't
-- already exist.
--
-- combatstate - The current state of combat
-- character - The character whose about to start swinging
-- ]]
function M.swinging.apply(character, combatstate)
    statusops.increaseattrition(combatstate, character, 2)
    local bonus = 2
    local attackmodifiers = character.secondarystatistics.attackmodifiers
    local defensemodifiers = character.secondarystatistics.defensemodifiers
    attackmodifiers[M.swinging.name] = bonus
    defensemodifiers[M.swinging.name] = bonus
    cancelstatuscommand(
        combatstate.choicecounts[character.name],
        "end swinging",
        M.swinging,
        character.name
    )
end

function M.swinging.remove(character, combatstate)
    statusops.increaseattrition(combatstate, character, -2)
    local attackmodifiers = character.secondarystatistics.attackmodifiers
    attackmodifiers[M.swinging.name] = nil
    local defensemodifiers = character.secondarystatistics.defensemodifiers
    defensemodifiers[M.swinging.name] = nil
    removechoice(combatstate.choicecounts[character.name], "end swinging")
end

function M.swinging.startturn(character, combatstate)
    local statuses = require 'data/statuses'
    local away = statuses.away
    if statusops.statusduration(character, away) then
        statusops.removestatus(character, away, combatstate)
        table.insert(
            combatstate.combatlog, 
            character.combatname .. " zips back into the fray! "
               .. character.combatname .. " is no longer Away!"
       )
    end
end

---------------------------- Scarlet Paddle -----------------------------------
--[[
-- Scarlet Paddle is a weapon that adds an additional action: Spank to the
-- character's repetoire. Spank inflicts Humiliated on a target. Humiliated
-- increases the spankee's strength by spanker.strength / 2, but decreases
-- speed and power by the same amount. It also can affect the abilities of
-- some enemies.
--
-- It will also reduce the duration of the target's best buff by duration.
--]]
M.scarletpaddle = {
    name="Scarlet Paddle",
    display="paddle",
    concisename="PA",
    buff=1,
    type="weapon"
}

function M.scarletpaddle.remove(character, combatstate)
    local choices = combatstate.choicecounts[character.name]
    removechoice(choices, "end paddle")
    removechoice(choices, "spank")
    statusops.increasechoice(choices, "attack", 1)
end

function M.scarletpaddle.apply(character, combatstate)
    local hasweapon = M.stackweapons(combatstate, character, M.scarletpaddle)
    if hasweapon then
        return
    end
    local choices = combatstate.choicecounts[character.name]
    statusops.increasechoice(choices, "spank", 1, scarletchoices.spank())
    statusops.decreasechoice(choices, "attack", 1)
    cancelstatuscommand(
        combatstate.choicecounts[character.name],
        "end paddle",
        M.scarletpaddle,
        character.name
    )
end

---------------------------- Scarlet Armor -----------------------------------
--[[
-- Scarlet Armor decreases the amount of damage the target takes by power/4.
--]]
M.armor = {
    name="Armor",
    display="armor",
    concisename="AR",
    buff=3
}

function M.armor.apply(character, combatstate)
    local name = character.name
    statusops.increaseattrition(combatstate, character, 1)
    local buff = math.ceil(character.statistics.power / 4)
    character.secondarystatistics.defensemodifiers[M.armor.name] = buff
    cancelstatuscommand(
        combatstate.choicecounts[character.name],
        "end armor",
        M.armor,
        character.name
    )
end

function M.armor.remove(character, combatstate)
    statusops.increaseattrition(combatstate, character, -1)
    character.secondarystatistics.defensemodifiers[M.armor.name] = nil
    removechoice(combatstate.choicecounts[character.name], "end armor")
end
----------------------- End Armor -------------------------------

----------------------- Begin Scarlet Focus -------------------------------
-- Scarlet Focus increases the player's power by 1, but decreases her attack by
-- 1.

M.scarletfocus = {
    name="scarletfocus",
    display="Focus",
    concisename="F"
}
function M.scarletfocus.apply(character, combatstate, duration)
    local choices = combatstate.choicecounts[character.name]
    cancelstatuscommand(choices, "end focus", M.scarletfocus, character.name)
    choices.attack.current = choices.attack.current - 1
    choices.attack.max = choices.attack.max - 1
    choices.power.current = choices.power.current + 1
    choices.power.max = choices.power.max + 1
end

function M.scarletfocus.remove(character, combatstate)
    local choices = combatstate.choicecounts[character.name]
    removechoice(choices, "end focus")
    choices.attack.current = choices.attack.current + 1
    choices.attack.max = choices.attack.max + 1
    choices.power.current = choices.power.current - 1
    choices.power.max = choices.power.max - 1
end
----------------------- End Scarlet Focus -------------------------------

return M
