-- [[
-- Copyright 2019 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- A simple script that generates a giant org file with the dynamic text
-- resolved to specific values. This allows us to generate a plain text,
-- HTML version of the game for people who can't run the game itself for
-- whatever weird reason.
--]]

local M = {}
local characters = require 'data/characters'
local items = require 'data/items'
local load = require 'story/load'
local state = require 'state'

local function writeEvents(resolvedevents)
    local scarlettables = require 'scarlettables'
    local eventcontent = {}
    if resolvedevents.start then
        local start = resolvedevents.start
        resolvedevents.start = nil
        table.insert(eventcontent, "* start")
        table.insert(eventcontent, start.body)
        if start.transformer then
            table.insert(eventcontent, "** do")
            table.insert(eventcontent, start.transformer)
        end
        if start.automatics then
            table.insert(eventcontent, "** automatics")
            table.insert(eventcontent, start.automatics)
        end
        if start.choices then
            table.insert(eventcontent, "** choices")
            table.insert(eventcontent, start.choices)
        end
    end
    table.insert(eventcontent, "")
    for name, event in pairs(resolvedevents) do
        table.insert(eventcontent, "* " .. name)
        if event.allies then
            table.insert(eventcontent, "** allies")
            table.insert(eventcontent, event.allies)
            table.insert(eventcontent, "** enemies")
            table.insert(eventcontent, event.enemies)
            table.insert(eventcontent, "** victory")
            table.insert(eventcontent, event.victory)
            if event.defeat then
                table.insert(eventcontent, "** defeat")
                table.insert(eventcontent, event.defeat)
            end
        else
          table.insert(eventcontent, event.body)
          if event.transformer then
              table.insert(eventcontent, "** do")
              table.insert(eventcontent, event.transformer)
          end
          if event.automatics then
              table.insert(eventcontent, "** automatics")
              table.insert(eventcontent, event.automatics)
          end
          if event.choices then
              table.insert(eventcontent, "** choices")
              table.insert(eventcontent, event.choices)
          end
          table.insert(eventcontent, "")
        end
    end
    local resolvedtext = assert(io.open("data/story/scarlet_moon.org", "w"))
    resolvedtext:write(table.concat(eventcontent, "\n"))
    resolvedtext:close()
end

function M.generateOrgMode()
    local configuration = state.buildconfiguration()
    local stories = configuration.stories
    local samplepc = {}
    for key, value in pairs(characters.pc) do
        samplepc[key] = value
    end
    samplepc.attributes.gender = GENERATE_HTML_GENDER
    if samplepc.attributes.gender == "male" then
        samplepc.printedname = "Adrian"
    end
    samplepc.clothing = {_keys=samplepc.clothing._keys}
    samplepc.clothing.upper = assert(items["t-shirt"])
    samplepc.clothing.lower = assert(items.jeans)
    samplepc.clothing.underwear = samplepc.attributes.gender == "female" 
        and assert(items.thong) 
        or assert(items["boxer briefs"])
    local attributescopy = {}
    for name, value in pairs(samplepc.attributes) do
        attributescopy[name] = value
    end
    attributescopy.skincolor = "peach"
    attributescopy.bodytype = "slender"
    samplepc.attributes = attributescopy
    local gamestate = state.buildgamestate(configuration)
    gamestate.keywords = {placeholder=1}
    gamestate.characters.pc = samplepc
    local resolvedevents = {}
    local skipmarker = "$$skip$$"
    for name, event in pairs(configuration.stories) do
        if string.sub(name, 1, #skipmarker) ~= skipmarker then
            local resolvedevent = {}
            if event.allies then
                local allies = {}
                for i, name in ipairs(event.allies) do
                    allies[i] = "- " .. name
                end
                resolvedevent.allies = table.concat(allies, "\n")
                local enemies = {}
                for i, name in ipairs(event.enemies) do
                    enemies[i] = "- " .. name
                end
                resolvedevent.enemies = table.concat(enemies, "\n")
                resolvedevent.victory = "[[#" .. string.gsub(event.victory, "_", "-") .. "][Victory!]]"
                if event.defeat then
                    resolvedevent.defeat = "[[#" .. string.gsub(event.defeat, "_", "-") .. "][Oh noes, you've been defeated!]]"
                end
            else
                resolvedevent.body = table.concat(event.body(gamestate), "\n\n")
                if event.choices then
                  local choices = event.choices(gamestate)
                  local resolvedchoices = {}
                  for _, choice in ipairs(choices._keys) do
                      local nexteventname = string.gsub(choices[choice].name, "_", "-")
                      table.insert(
                          resolvedchoices,
                          "- [[#" .. nexteventname .. "][" .. choice .. "]]"
                      )
                  end
                  if next(resolvedchoices) then
                    resolvedevent.choices = table.concat(resolvedchoices, "\n")
                  end
                end
                local automatics = event.automatics and event.automatics(gamestate, true)
                if automatics then
                    local resolvedautomatics = {}
                    for _, predicate in ipairs(automatics._keys) do
                        local eventname = string.gsub(automatics[predicate], "_", "-")
                        table.insert(
                            resolvedautomatics,
                            "- [[#" .. eventname .. "][" .. predicate .. "]]"
                        )
                    end
                    resolvedevent.automatics = table.concat(resolvedautomatics, "\n")
                end
                local doblock = event.transformer and event.transformer(gamestate, true)
                if doblock then
                    local operations = {}
                    for _, operation in ipairs(doblock) do
                        table.insert(operations, "- " .. operation)
                    end
                    resolvedevent.transformer = table.concat(operations, "\n")
                end
            end
            resolvedevents[string.gsub(name, "_", "-")] = resolvedevent
        end
    end
    writeEvents(resolvedevents)
end

return M
