--[[
-- Copyright 2016 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- Contains a collection of utility functions that prove useful across
-- multiple UI screens, like pattern matching choices
--]]
local M = {}

--[[
-- Applies the "show a sublist of values that match a pattern" flow to
-- arbitrary sets of data.
--
-- valueswrapper: A table wrapping the list of values the player can choose from
--  It likely contains other metadata about the list of values, like the clothing
--  slot associated with those values, or the personalization id. It must have
--  a key `values` containing the list of values to choose from
-- command: The string provided by the player that we're using to filter the values
-- nameextractor: A function that takes a value from valueswrapper.values and returns
--  a string we can pattern match against
-- choosesublist: A function that takes a table with the same structure as 
-- valueswrapper and returns `true` if the player has made a selection, false
-- otherwise. This will often be a closure wrapping the function calling this
-- function, allowing us to recursively apply filters to lists of values
--]]
function M.patternmatch(valueswrapper, command, nameextractor, choosesublist)
    local escaped = string.gsub(command, "%-", "%%-")
    local pattern = ".*" .. escaped .. ".*"
    local sublist = {}
    for _, value in ipairs(valueswrapper.values) do
        if string.match(nameextractor(value), pattern) then
            table.insert(sublist, value)
        end
    end
    local scarlettables = require 'scarlettables'
    if #sublist == 0 then
        print("No matches found.")
        return choosesublist(valueswrapper)
    end
    local subvalueswrapper = {}
    for key, value in pairs(valueswrapper) do
        subvalueswrapper[key] = value
    end
    subvalueswrapper.values = sublist
    subvalueswrapper.search = command
    local updated = choosesublist(subvalueswrapper)
    if not updated then
        return choosesublist(valueswrapper)
    end
    return updated
end


return M
