--[[
-- Copyright 2016 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- fontERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

local M = {}
local love = require 'love'
local pages = require 'pages'

local bodyfonts = {}

--[[
-- A function that takes a font size, and returns a new font for body
-- text of the specified size.
--]]
function M.buildbody(fontsize)
    if not bodyfonts[fontsize] then
        bodyfonts[fontsize] = assert(
            love.graphics.newFont("ui/NewCenturySchoolbook.ttf", fontsize)
        )
    end
    return bodyfonts[fontsize]
end

local combatfonts = {}

function M.buildcombatboard(fontsize)
    if not combatfonts[fontsize] then
        combatfonts[fontsize] = assert(
            love.graphics.newFont("ui/Inconsolata.otf", fontsize)
        )
    end
    return combatfonts[fontsize]
end

local boldbuttonfonts = {}
local buttonfonts = {}

-- [[
-- A function that takes a font size, and whether or not the font should
-- be bold (defaults to false) and returns a new font for buttons.
-- ]]
function M.buildbutton(fontsize, bold)
    if bold then
        if not boldbuttonfonts[fontsize] then
            boldbuttonfonts[fontsize] = assert(
                love.graphics.newFont("ui/verdanab.ttf", fontsize)
            )
        end
        return boldbuttonfonts[fontsize] 
    else
        if not buttonfonts[fontsize] then
            buttonfonts[fontsize] = assert(
                love.graphics.newFont("ui/verdana.ttf", fontsize)
            )
        end
        return buttonfonts[fontsize] 
    end
end

--[[
-- Given a font, returns the maximum width of a single character in that
-- font.
--
-- The character used is `W` because that tends to be the widest character
-- in M.
--]]
function M.singleCharacterWidth(font)
    return font:getWidth('W')
end

--[[
-- Given a font, and some text, returns a width that will comfortably
-- fit the text without soft wraps.
--
-- In order to ensure some buffering, we compute the width needed by the
-- specified text, and tack on the width of a single character in the
-- specified font.
--]]
function M.width(font, text)
    return font:getWidth(text) + M.singleCharacterWidth(font)
end

M.BOLD = true

M.titlefontsize = 32
function M.title(titlefontsize)
    local fontsize fontsize = tonumber(titlefontsize or M.titlefontsize)
    return love.graphics.newFont("ui/verdana.ttf", fontsize)
end
function M.titleBold(titlefontsize)
    return love.graphics.newFont(
        "ui/verdanab.ttf",
        tonumber(titlefontsize or M.titlefontsize)
    )
end

M.buttonfontsize = 20
function M.button(buttonfontsize)
    local fontsize = tonumber(buttonfontsize or M.buttonfontsize)
    return M.buildbutton(fontsize)
end
function M.buttonBold(buttonfontsize)
    local fontsize = tonumber(buttonfontsize or M.buttonfontsize)
    return M.buildbutton(fontsize, M.BOLD)
end

M.smallbuttonfontsize = 16
function M.smallbutton(buttonfontsize)
    local fontsize = tonumber(buttonfontsize or M.smallbuttonfontsize)
    return M.buildbutton(fontsize)
end

M.bodyfontsize = 20
function M.body(bodyfontsize)
    local fontsize = tonumber(bodyfontsize or M.bodyfontsize)
    return M.buildbody(fontsize)
end

M.combatboardsize = 24
function M.combatboard(combatfontsize)
    local fontsize = tonumber(combatfontsize or M.combatboardsize)
    return M.buildcombatboard(fontsize)
end

local function changefontsizes(options, amount, fontname)
    if fontname then
        options[fontname] =
            assert(options[fontname], "Unknown font: " .. fontname) + amount
    else
        local fontnames = {"titlefont", "bodyfont", "combatfont", "buttonfont"}
        for _, name in pairs(fontnames) do
            local fontsize = tonumber(options[name])
            local newfontsize = fontsize + amount
            if 8 <= newfontsize and newfontsize <= 32 then
                options[name] = tostring(newfontsize)
            end
        end
    end
end


--[[
-- Returns a page that increases the size of the specified font by 1. If no
-- font is specified, then increases all M by 1
--
-- previouspage: The page to transition to after this
-- fontname: The name of the font to increase, if nil increase all M
-- ignoreshift: A boolean indicating whether or not to check for shift. Some
--  keyboards (i.e. Danish keyboards) have a literal "+" key, so we don't want
--  to check for shift on those keyboards.
--]]
function M.increasefontsizepage(previouspage, fontname, ignoreshift)
    return pages.page(
        nil,
        function() return previouspage end,
        nil,
        function(uistate)
            local shiftheld = ignoreshift
                or love.keyboard.isDown("lshift")
                or love.keyboard.isDown("rshift")
            if shiftheld then
                changefontsizes(uistate.core.gamestate.options, 1, fontname)
            end
        end
    )
end

--[[
-- Returns a page that Decreases the size of the specified font by 1. If no
-- font is specified, then decrease all M by 1
--
-- previouspage: The page to transition to after this
-- fontname: The name of the font to increase, if nil increase all M
--]]
function M.decreasefontsizepage(previouspage, fontname)
    return pages.page(
        nil,
        function() return previouspage end,
        nil,
        function(uistate)
            changefontsizes(uistate.core.gamestate.options, -1, fontname)
        end
    )
end

return M
