--[[
    Copyright (C) 2020  AKA Russell

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- A simple script that takes in all of Scarlet Moon's story files and verifies
-- that:
-- 1. All their automatics and choice links exist.
-- 2. There aren't any duplicate events.
-- 3. They all have children events (assuming they're not marked with the leaf marker)
--]]

local F = require 'f-strings/F'
local orgparser = require 'orgparser'

local allevents = {}
local errors = {}
for storyfilename in io.lines("story_files") do
    local events = orgparser.decode(io.lines(storyfilename), storyfilename)
    for _, name in ipairs(events._keys) do
        local event = events[name]
        if allevents[name] then
            table.insert(errors, F"Event {name} already exists.")
        else
            allevents[name] = event
        end
    end
end

for name, event in pairs(allevents) do
    for _, subcategory in ipairs({"choices", "automatics"}) do
        if event[subcategory] then
            for _, key in pairs(event[subcategory]._keys) do
                local eventname = event[subcategory][key]
                if not allevents[eventname] then
                    table.insert(
                        errors, 
                        F"Unknown event '{eventname}' in '{name}'."
                    )
                end
            end
        end 
    end
    if event.victory and not allevents[event.victory] then
        table.insert(
            errors,
            F"Unknown event '{event.victory}' in '{name}'"
        )
    end
    if event.defeat and not allevents[event.defeat] then
        table.insert(
            errors,
            F"Unknown event '{event.defeat}' in '{name}'"
        )
    end
    local missingchildren = not next(event.choices or {}) 
        and not next(event.automatics or {}) 
        and not event.victory
        and not event.defeat
        and not event.leaf 
    if missingchildren then
        table.insert(
            errors,
            F"Event {name} has no children and is not a leaf."
        )
    end
end

if #errors > 0 then
    print(table.concat(errors, "\n"))
    return 1
else
    print("All good!")
end
